﻿using System;
using System.Threading;
using System.Threading.Tasks;
using System.Threading.Tasks.Dataflow;

namespace KcpNet
{
    internal class KcpByteBuffer
    {
        private readonly BufferBlock<byte[]> _bufferQueue;

        private int _bufferOffset;
        private byte[] _buffer;


        public KcpByteBuffer()
        {
            _bufferQueue = new BufferBlock<byte[]>();
        }

        public void Write(byte[] bytes)
        {
            if (!_bufferQueue.Post(bytes))
            {
                throw new InvalidOperationException("Write bytes error.");
            }
        }

        public async Task WriteAsync(byte[] bytes)
        {
            if (!await _bufferQueue.SendAsync(bytes).ConfigureAwait(false))
            {
                throw new InvalidOperationException("Write bytes error.");
            }
        }


        public async Task<int> ReadAsync(byte[] buffer, CancellationToken cancellationToken)
        {
            try
            {
                if (_buffer != null)
                {
                    var bufferLength = _buffer.Length - _bufferOffset;
                    if (buffer.Length >= bufferLength)
                    {
                        var length = bufferLength;
                        Array.Copy(_buffer, _bufferOffset, buffer, 0, length);
                        _buffer = null;
                        _bufferOffset = 0;
                        return length;
                    }

                    Array.Copy(_buffer, _bufferOffset, buffer, 0, buffer.Length);
                    _bufferOffset += buffer.Length;
                    return buffer.Length;

                }

                _buffer = await _bufferQueue.ReceiveAsync(cancellationToken).ConfigureAwait(false);
                if (_buffer != null)
                {
                    if (_buffer.Length <= buffer.Length)
                    {
                        var length = _buffer.Length;
                        Array.Copy(_buffer, buffer, length);
                        _buffer = null;
                        return length;
                    }

                    if (_buffer.Length > buffer.Length)
                    {
                        Array.Copy(_buffer, buffer, buffer.Length);
                        _bufferOffset += buffer.Length;
                        return buffer.Length;
                    }
                }
            }
            catch (Exception)
            {
                //This could be the connection is closed.
            }

            return 0;
        }


        public async Task<int> ReadAsync(byte[] buffer)
        {
            return await ReadAsync(buffer, CancellationToken.None).ConfigureAwait(false);
        }

        public int Read(byte[] buffer)
        {
            try
            {
                if (_buffer != null)
                {
                    var bufferLength = _buffer.Length - _bufferOffset;
                    if (buffer.Length >= bufferLength)
                    {
                        var length = bufferLength;
                        Array.Copy(_buffer, _bufferOffset, buffer, 0,length);
                        _buffer = null;
                        _bufferOffset = 0;
                        return length;
                    }

                    Array.Copy(_buffer, _bufferOffset, buffer,0, buffer.Length);
                    _bufferOffset += buffer.Length;
                    return buffer.Length;

                }

                _buffer = _bufferQueue.Receive();
                if (_buffer != null)
                {
                    if (_buffer.Length <= buffer.Length)
                    {
                        var length = _buffer.Length;
                        Array.Copy(_buffer, buffer, length);
                        _buffer = null;
                        return length;
                    }

                    if (_buffer.Length > buffer.Length)
                    {
                        Array.Copy(_buffer, buffer, buffer.Length);
                        _bufferOffset += buffer.Length;
                        return buffer.Length;
                    }
                }
            }
            catch (Exception)
            {
                //This could be the connection is closed.
            }

            return 0;
        }

        public void Close()
        {
            _bufferQueue.TryReceiveAll(out _);
            _bufferQueue.Complete();
            _buffer = null;
            _bufferOffset = 0;
        }
    }
}
