﻿using System;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;

namespace KcpNet
{
    public class KcpClient : KcpNetBase
    {
        private readonly UdpClient _client;

        private bool _connected;

        private IPEndPoint _remoteEndPoint;

        private KcpConnection _connection;

        private Thread _processThread;
        private Thread _updateTread;
        private Thread _autoPingThread;

        public TimeSpan ConnectionTimeout { get; set; } = TimeSpan.FromSeconds(5);

        public TimeSpan PingInterval { get; set; } = TimeSpan.FromSeconds(2);

        public bool AutoPing { get; set; } = true;

        public KcpConnection Connection => _connection;

        public KcpClient()
        {
            _client = new UdpClient();
        }

        public void Connect(IPEndPoint remoteEndPoint, int timeout = Timeout.Infinite)
        {
            _remoteEndPoint = remoteEndPoint;
            var handshakeTask = HandshakeAsync();
            var timeoutTask = Task.Delay(timeout);
            var tasks = new[] { handshakeTask, timeoutTask };
            var resultTask = Task.WhenAny(tasks).Result;
            if (resultTask == handshakeTask)
            {
                _connection = handshakeTask.Result;
                if (_connection != null)
                {
                    _connection.Closed += OnConnectionClosed;
                    _connected = true;
                    _processThread = new Thread(() =>
                    {
                        while (_connected)
                        {
                            try
                            {
                                _connection.ProcessOnce();
                                Thread.Sleep(1);
                            }
                            catch (Exception)
                            {
                                _connection.Close();
                                break;
                            }

                        }
                    });
                    _updateTread = new Thread(() =>
                    {
                        while (_connected)
                        {
                            try
                            {
                                _connection.Update();
                                var now = DateTime.UtcNow;
                                if (now - _connection.LastPing > ConnectionTimeout)
                                {
                                    _connection.Close();
                                }

                                Thread.Sleep(10);
                            }
                            catch (Exception)
                            {
                                _connection.Close();
                                break;
                            }

                        }
                    });
                    _processThread.Start();
                    _updateTread.Start();

                    if (AutoPing)
                    {
                        _autoPingThread = new Thread(() =>
                        {
                            while (_connected)
                            {
                                try
                                {
                                    _connection.PingRequest();
                                    Thread.Sleep(PingInterval);
                                }
                                catch (Exception)
                                {
                                    _connection.Close();
                                    break;
                                }
                            }

                        });
                        _autoPingThread.Start();
                    }

                    Task.Run(async () =>
                    {
                        while (_connected)
                        {
                            var result = await _client.ReceiveAsync().ConfigureAwait(false);
                            if (result.RemoteEndPoint.Equals(_remoteEndPoint))
                            {
                                ((IKcpIoContainer)_connection).Io.Input(result.Buffer);
                            }
                        }
                    });
                }
                else
                {
                    throw new InvalidOperationException("Connect to server failed.");
                }
            }
            else
            {
                _client.Close();
                _client.Dispose();
                throw new TimeoutException("Connect to server timeout.");
            }
        }

        private void OnConnectionClosed(object sender, EventArgs e)
        {
            _connection.Closed -= OnConnectionClosed;
            _connected = false;
        }

        public void Close()
        {
            if (_connected)
            {
                _connection?.Close();
                _processThread?.Join();
                _updateTread?.Join();
                _autoPingThread?.Join();
            }
        }


        public async Task SendAsync(byte[] buffer)
        {
            await SendAsync(buffer, CancellationToken.None).ConfigureAwait(false);
        }

        public async Task SendAsync(byte[] buffer, CancellationToken cancellationToken)
        {
            if (_connection != null)
            {
                await _connection.SendAsync(buffer, cancellationToken).ConfigureAwait(false);
            }
            else
            {
                throw new NullReferenceException("Connection is null");
            }
        }


        public async Task<int> ReadAsync(byte[] buffer)
        {
            return await ReadAsync(buffer, CancellationToken.None).ConfigureAwait(false);
        }

        public async Task<int> ReadAsync(byte[] buffer, CancellationToken cancellationToken)
        {
            if (_connection != null)
            {
                return await _connection.ReadAsync(buffer,cancellationToken).ConfigureAwait(false);
            }

            throw new NullReferenceException("Connection is null");
        }


        private async Task<KcpConnection> HandshakeAsync()
        {
            try
            {
                var random = new Random();
                var sb01 = (byte) random.Next(0, 255);
                var sb02 = (byte) random.Next(0, 255);
                var su03 = (ushort) (sb01 * sb02);
                var su03Buffer = BitConverter.GetBytes(su03);
                var sendBuffer = new byte[8];
                Array.Copy(ProtocolHeader, 0, sendBuffer, 0, 3);
                sendBuffer[3] = sb01;
                sendBuffer[4] = sb02;
                Array.Copy(su03Buffer, 0, sendBuffer, 5, 2);
                sendBuffer[7] = (byte) HandshakeLevel.S01;

                await _client.SendAsync(sendBuffer, sendBuffer.Length, _remoteEndPoint).ConfigureAwait(false);

                var result = await _client.ReceiveAsync().ConfigureAwait(false);
                if (result.RemoteEndPoint.Equals(_remoteEndPoint))
                {
                    await using var stream = new MemoryStream(result.Buffer);
                    var header = await ReadBytesAsync(stream, 3).ConfigureAwait(false);
                    if (!header.SequenceEqual(ProtocolHeader))
                    {
                        throw new InvalidDataException("Error header for handshake01.");
                    }

                    var rb1 = await ReadByteAsync(stream).ConfigureAwait(false);
                    var rb2 = await ReadByteAsync(stream).ConfigureAwait(false);
                    var ru = await ReadUInt16Async(stream).ConfigureAwait(false);
                    if (rb1 * rb2 != ru)
                    {
                        throw new InvalidDataException("Error RBU data for handshake.");
                    }

                    var level = (HandshakeLevel) await ReadByteAsync(stream).ConfigureAwait(false);
                    if (level != HandshakeLevel.S01)
                    {
                        throw new InvalidDataException("Error level data for handshake.");
                    }

                    var connectionId = await ReadUInt32Async(stream).ConfigureAwait(false);

                    sendBuffer = new byte[12];
                    var sb11 = (byte) random.Next(0, 255);
                    var sb12 = (byte) random.Next(0, 255);
                    var su13 = (ushort) (sb11 * sb12);
                    var su13Buffer = BitConverter.GetBytes(su13);
                    Array.Copy(ProtocolHeader, 0, sendBuffer, 0, 3);
                    sendBuffer[3] = sb11;
                    sendBuffer[4] = sb12;
                    Array.Copy(su13Buffer, 0, sendBuffer, 5, 2);
                    sendBuffer[7] = (byte) HandshakeLevel.S02;
                    var idData = BitConverter.GetBytes(connectionId);
                    Array.Copy(idData, 0, sendBuffer, 8, 4);
                    await _client.SendAsync(sendBuffer, sendBuffer.Length, _remoteEndPoint).ConfigureAwait(false);
                    return new KcpConnection(connectionId, new KcpUdpIo(_client, _remoteEndPoint),(IPEndPoint) _client.Client.LocalEndPoint, _remoteEndPoint);
                }
            }
            catch
            {
                //
            }
            return null;
        }
        
    }
}
