﻿using System;
using System.IO;
using System.Net;
using System.Threading;
using System.Threading.Tasks;

namespace KcpNet
{
    public class KcpConnection : IKcpIoContainer
    {
        private static readonly IRentable MemoryPool = new KcpMemoryPool();

        private readonly IKcpIo _io;
        private readonly Kcp _kcp;


        private readonly Thread _readPacketThread;
        private readonly KcpByteBuffer _receivedByteBuffer;
        private readonly KcpDataBuffer _receivedDataBuffer;
        private readonly KcpPriorityPacketBuffer _sendPacketBuffer;

        private bool _closed;

        IKcpIo IKcpIoContainer.Io => _io;

        public uint Id { get; }

        public IPEndPoint LocalEndPoint { get; }


        public IPEndPoint RemoteEndPoint { get; }

        public TimeSpan SendTimeout { get; set; } = TimeSpan.FromSeconds(5);

        public TimeSpan ReceiveTimeout { get; set; } = TimeSpan.FromSeconds(5);


        public DateTime LastPing { get; private set; }


        public event EventHandler PingRequested;


        public event EventHandler PingResponded;


        public event EventHandler Closed;


        internal KcpConnection(uint connectionId, IKcpIo io, IPEndPoint localEndPoint, IPEndPoint remoteEndPoint)
        {
            LastPing = DateTime.UtcNow;
            Id = connectionId;
            _io = io;
            _io.Received += OnIoReceived;
            LocalEndPoint = localEndPoint;
            RemoteEndPoint = remoteEndPoint;
            _receivedByteBuffer = new KcpByteBuffer();
            _receivedDataBuffer = new KcpDataBuffer();
            _sendPacketBuffer = new KcpPriorityPacketBuffer(1,12,3);
            var callback = new KcpOutputCallback(_io);
            _kcp = new Kcp(Id, callback, MemoryPool);
            _kcp.NoDelay(1, 10, 2, 1);
            _kcp.WndSize(64, 64);
            _kcp.SetMtu(512);
            _closed = false;
            _readPacketThread = new Thread(() =>
            {
                try
                {
                    while (!_closed)
                    {
                        ReadPacketOnce();
                        Thread.Sleep(1);
                    }
                }
                catch (Exception)
                {
                    //Exception will be catch when the connection is closed.
                }
            });
            _readPacketThread.Start();
        }



        private void ProcessReceive()
        {
            int size;
            while ((size = _kcp.PeekSize()) > 0)
            {
                var buffer = new byte[size];
                var result = _kcp.Recv(buffer);
                if ( result >= 0)
                {
                    _receivedByteBuffer.Write(buffer);
                }
                else
                {
                    throw new InvalidOperationException($"Receive kcp message error:{result}.");
                }
            }
        }


        private void ProcessSend()
        {
            var packets = _sendPacketBuffer.ReadPackets();
            foreach (var packet in packets)
            {
                var result = _kcp.Send(packet.GetBuffer());
                if (result < 0)
                {
                    throw new InvalidOperationException($"Send kcp data error: {result}");
                }
            }
        }



        internal void ProcessOnce()
        {
            if (_closed) throw new InvalidOperationException("Connection closed.");
            ProcessReceive();
            ProcessSend();
        }


        /// <summary>
        /// This method should run in a standalone thread.
        /// </summary>
        private void ReadPacketOnce()
        {
            var typeData = new byte[1];
            var result = _receivedByteBuffer.Read(typeData);
            if (result == 0)
            {
                throw new InvalidOperationException("Receive packet type error.");
            }

            var type = (KcpPacketType) typeData[0];
            switch (type)
            {
                case KcpPacketType.PingRequest:
                    PingResponse();
                    LastPing = DateTime.UtcNow;
                    PingRequested?.Invoke(this, EventArgs.Empty);
                    break;
                case KcpPacketType.PingResponse:
                    LastPing = DateTime.UtcNow;
                    PingResponded?.Invoke(this, EventArgs.Empty);
                    break;
                case KcpPacketType.Close:
                    Close();
                    break;
                case KcpPacketType.Data:
                    var lengthData = new byte[4];
                    result = _receivedByteBuffer.Read(lengthData);
                    if (result == 0)
                    {
                        throw new InvalidOperationException("Receive length data error.");
                    }
                    var length = BitConverter.ToInt32(lengthData);
                    var data = new byte[length];
                    result = _receivedByteBuffer.Read(data);
                    if (result == 0)
                    {
                        throw new InvalidOperationException("Receive data error.");
                    }

                    var kcpData = new KcpData(data);
                    _receivedDataBuffer.Write(kcpData);
                    break;
                default:
                    throw new InvalidDataException("Unknown packet type.");
            }
        }


        internal void Update()
        {
            if (_closed) throw new InvalidOperationException("Connection closed.");
            _kcp.Update(DateTime.UtcNow);
        }


        private void OnIoReceived(object sender, byte[] e)
        {
            _kcp.Input(e);
        }

        public async Task SendAsync(byte[] buffer)
        {
            await SendAsync(buffer, CancellationToken.None).ConfigureAwait(false);
        }

        public async Task SendAsync(byte[] buffer, CancellationToken cancellationToken)
        {
            if (_closed) throw new InvalidOperationException("Connection closed.");
            var data = new KcpData(buffer);
            var waitTime = TimeSpan.Zero;
            while (true)
            {
                if (waitTime > SendTimeout)
                {
                    throw new TimeoutException("Send buffer timeout.");
                }
                if (await _sendPacketBuffer.WritePacketAsync(data, PacketPriority.Middle, cancellationToken))
                {
                    break;
                }

                await Task.Delay(100, cancellationToken).ConfigureAwait(false);
                waitTime = waitTime.Add(TimeSpan.FromMilliseconds(100));
            }
            
        }

        private void SendClose()
        {
            if (_closed) throw new InvalidOperationException("Connection closed.");
            _kcp.Input(KcpClose.Shared.GetBuffer());
            _kcp.Update(DateTime.UtcNow);
        }

        internal void PingResponse()
        {
            if (_closed) throw new InvalidOperationException("Connection closed.");
            var waitTime = TimeSpan.Zero;
            while (true)
            {
                if (waitTime > SendTimeout)
                {
                    throw new TimeoutException("Ping response timeout.");
                }
                if (_sendPacketBuffer.WritePacket(KcpPingResponse.Shared, PacketPriority.High))
                {
                    break;
                }
                Task.Delay(10).Wait();
                waitTime = waitTime.Add(TimeSpan.FromMilliseconds(10));
            }
        }

        internal void PingRequest()
        {
            if (_closed) throw new InvalidOperationException("Connection closed.");
            var waitTime = TimeSpan.Zero;
            while (true)
            {
                if (waitTime > SendTimeout)
                {
                    throw new TimeoutException("Ping request timeout.");
                }
                if (_sendPacketBuffer.WritePacket(KcpPingRequest.Shared, PacketPriority.High))
                {
                    break;
                }
                Task.Delay(10).Wait();
                waitTime = waitTime.Add(TimeSpan.FromMilliseconds(10));
            }
        }

        public async Task<int> ReadAsync(byte[] buffer)
        {
            return await ReadAsync(buffer, CancellationToken.None).ConfigureAwait(false);
        }

        public async Task<int> ReadAsync(byte[] buffer, CancellationToken cancellationToken)
        {
            if (_closed) throw new InvalidOperationException("Connection closed.");
            var timeoutTask = Task.Delay(ReceiveTimeout,cancellationToken);
            var readTask = _receivedDataBuffer.ReadAsync(buffer, cancellationToken);
            var resultTask = await Task.WhenAny(new[]{timeoutTask, readTask}).ConfigureAwait(false);
            if (resultTask == timeoutTask)
            {
                throw new TimeoutException("Read buffer timeout.");
            }
            return readTask.Result;
        }


        public void Close()
        {
            if (!_closed)
            {
                SendClose();
                _closed = true;
                _receivedDataBuffer.Close();
                _receivedByteBuffer.Close();
                _sendPacketBuffer.Close();
                _readPacketThread.Join();
                _io.Received -= OnIoReceived;
                _kcp.Dispose();
                Closed?.Invoke(this, EventArgs.Empty);
            }
        }
    }
}
