﻿using System;
using System.Threading;
using System.Threading.Tasks;
using System.Threading.Tasks.Dataflow;

namespace KcpNet
{
    internal class KcpDataBuffer
    {
        private readonly BufferBlock<KcpData> _bufferQueue;

        private KcpData _buffer;

        public KcpDataBuffer()
        {
            _bufferQueue = new BufferBlock<KcpData>();
        }

        public void Write(KcpData data)
        {
            if (!_bufferQueue.Post(data))
            {
                throw new InvalidOperationException("Write data error.");
            }
        }

        public async Task WriteAsync(KcpData data)
        {
            if (!await _bufferQueue.SendAsync(data).ConfigureAwait(false))
            {
                throw new InvalidOperationException("Write data error.");
            }
        }


        public async Task<int> ReadAsync(byte[] buffer, CancellationToken cancellationToken)
        {
            try
            {
                if (_buffer != null)
                {
                    if (buffer.Length >= _buffer.Length)
                    {
                        var length = _buffer.Length;
                        Array.Copy(_buffer.GetData(), buffer, length);
                        _buffer = null;
                        return length;
                    }

                    Array.Copy(_buffer.GetData(), buffer, buffer.Length);
                    _buffer.Offset += buffer.Length;
                    return buffer.Length;

                }

                _buffer = await _bufferQueue.ReceiveAsync(cancellationToken).ConfigureAwait(false);
                if (_buffer != null)
                {
                    if (_buffer.Length <= buffer.Length)
                    {
                        var length = _buffer.Length;
                        Array.Copy(_buffer.GetData(), buffer, length);
                        _buffer = null;
                        return length;
                    }

                    if (_buffer.Length > buffer.Length)
                    {
                        Array.Copy(_buffer.GetData(), buffer, buffer.Length);
                        _buffer.Offset += buffer.Length;
                        return buffer.Length;
                    }
                }
            }
            catch (Exception)
            {
                //This could be the connection is closed.
            }

            return 0;
        }


        public async Task<int> ReadAsync(byte[] buffer)
        {
            return await ReadAsync(buffer, CancellationToken.None).ConfigureAwait(false);
        }


        public int Read(byte[] buffer)
        {
            try
            {
                if (_buffer != null)
                {
                    if (buffer.Length >= _buffer.Length)
                    {
                        var length = _buffer.Length;
                        Array.Copy(_buffer.GetData(), buffer, length);
                        _buffer = null;
                        return length;
                    }

                    Array.Copy(_buffer.GetData(), buffer, buffer.Length);
                    _buffer.Offset += buffer.Length;
                    return buffer.Length;

                }

                _buffer = _bufferQueue.Receive();
                if (_buffer != null)
                {
                    if (_buffer.Length <= buffer.Length)
                    {
                        var length = _buffer.Length;
                        Array.Copy(_buffer.GetData(), buffer, length);
                        _buffer = null;
                        return length;
                    }

                    if (_buffer.Length > buffer.Length)
                    {
                        Array.Copy(_buffer.GetData(), buffer, buffer.Length);
                        _buffer.Offset += buffer.Length;
                        return buffer.Length;
                    }
                }
            }
            catch (Exception)
            {
                //This could be the connection is closed.
            }

            return 0;
        }


        public void Close()
        {
            _bufferQueue.TryReceiveAll(out _);
            _bufferQueue.Complete();
            _buffer = null;
        }
    }
}
