﻿using System;
using System.Collections.Concurrent;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;
using System.Threading.Tasks.Dataflow;

namespace KcpNet
{
    public class KcpListener : KcpNetBase
    {
        private uint _id;

        private Thread _processThread;
        private Thread _updateThread;

        private readonly IPEndPoint _localEndPoint;

        private readonly BufferBlock<KcpConnection> _incomeConnections = new();

        private readonly ConcurrentDictionary<EndPoint, KcpConnection> _connections = new();

        private UdpClient _client;

        private bool _closed;


        public TimeSpan ConnectionTimeout { get; set; } = TimeSpan.FromSeconds(8);

        public KcpListener(IPEndPoint localEndPoint)
        {
            _localEndPoint = localEndPoint;
        }

        public async Task<KcpConnection> GetConnectionAsync()
        {
            return await GetConnectionAsync(CancellationToken.None).ConfigureAwait(false);
        }

        public async Task<KcpConnection> GetConnectionAsync(CancellationToken cancellationToken)
        {
            try
            {
                var connection = await _incomeConnections.ReceiveAsync(cancellationToken).ConfigureAwait(false);
                return connection;
            }
            catch
            {
                //
            }

            return null;
        }

        public void Start()
        {
            if (_closed)
            {
                throw new InvalidOperationException("Listener already closed.");
            }
            _client = new UdpClient(_localEndPoint);
            _processThread = new Thread(() =>
            {
                while (!_closed)
                {
                    var connections = _connections.Values;
                    Parallel.ForEach(connections, connection =>
                    {
                        try
                        {
                            connection.ProcessOnce();
                        }
                        catch (Exception)
                        {
                            connection.Close();
                        }
                    });
                    Thread.Sleep(10);
                }
            });
            _updateThread = new Thread(() =>
            {
                while (!_closed)
                {
                    var connections = _connections.Values;
                    Parallel.ForEach(connections, connection =>
                    {
                        try
                        {
                            connection.Update();
                            var now = DateTime.UtcNow;
                            if (now - connection.LastPing > ConnectionTimeout)
                            {
                                connection.Close();
                            }
                        }
                        catch (Exception)
                        {
                            connection.Close();
                        }
                    });
                    Thread.Sleep(10);
                }
            });
            _processThread.Start();
            _updateThread.Start();

            Task.Run(async () =>
                {
                    while (!_closed)
                    {
                        try
                        {
                            var result = await _client.ReceiveAsync().ConfigureAwait(false);
                            if (_connections.TryGetValue(result.RemoteEndPoint, out var connection))
                            {
                                ((IKcpIoContainer) connection).Io.Input(result.Buffer);
                            }
                            else
                            {
                                connection = await HandshakeAsync(result.RemoteEndPoint, result.Buffer, result.Buffer.Length).ConfigureAwait(false);
                                if (connection != null)
                                {
                                    connection.Closed += OnConnectionClosed;
                                    _connections.AddOrUpdate(result.RemoteEndPoint, _ => connection, (_, exist) =>
                                    {
                                        exist.Close();
                                        return connection;
                                    });
                                    await _incomeConnections.SendAsync(connection).ConfigureAwait(false);
                                }
                            }
                        }
                        catch (Exception)
                        {
                           //Could be the socket closed. 
                        }
                    }
                }
            );
        }

        private void OnConnectionClosed(object sender, EventArgs e)
        {
            var connection = (KcpConnection) sender;
            connection.Closed -= OnConnectionClosed;
            _connections.TryRemove(connection.RemoteEndPoint, out _);
        }


        private async Task<KcpConnection> HandshakeAsync(IPEndPoint remoteEndPoint, byte[] buffer, int bufferSize)
        {
            var handshakeBuffer = new byte[bufferSize];
            Array.Copy(buffer,handshakeBuffer, bufferSize);
            await using var stream = new MemoryStream(handshakeBuffer);
            var random = new Random();
            var header = await ReadBytesAsync(stream, 3).ConfigureAwait(false);
            if (!header.SequenceEqual(ProtocolHeader))
            {
                throw new InvalidDataException("Error header for handshake01.");
            }

            var rb1 = await ReadByteAsync(stream).ConfigureAwait(false);
            var rb2 = await ReadByteAsync(stream).ConfigureAwait(false);
            var ru = await ReadUInt16Async(stream).ConfigureAwait(false);
            if (rb1 * rb2 != ru)
            {
                throw new InvalidDataException("Error RBU data for handshake.");
            }

            var level = (HandshakeLevel)await ReadByteAsync(stream).ConfigureAwait(false);

            if (level == HandshakeLevel.S01)
            {
                var sb01 = (byte)random.Next(0, 255);
                var sb02 = (byte)random.Next(0, 255);
                var su03 = (ushort)(sb01 * sb02);
                var su03Buffer = BitConverter.GetBytes(su03);
                var sendBuffer = new byte[12];
                Array.Copy(ProtocolHeader, 0, sendBuffer, 0, 3);
                sendBuffer[3] = sb01;
                sendBuffer[4] = sb02;
                Array.Copy(su03Buffer, 0, sendBuffer, 5, 2);
                sendBuffer[7] = (byte)HandshakeLevel.S01;
                var connectionId = Interlocked.Increment(ref _id);
                var idData = BitConverter.GetBytes(connectionId);
                Array.Copy(idData, 0, sendBuffer, 8, 4);
                await _client.SendAsync(sendBuffer, sendBuffer.Length, remoteEndPoint);
            }
            else if (level == HandshakeLevel.S02)
            {
                var connectionId = await ReadUInt32Async(stream).ConfigureAwait(false);
                var connection = new KcpConnection(connectionId, new KcpUdpIo(_client, remoteEndPoint), (IPEndPoint)_client.Client.LocalEndPoint, remoteEndPoint);
                return connection;
            }
            else
            {
                throw new InvalidDataException("Error level data for handshake.");
            }

            return null;
        }


        public void Close()
        {
            _closed = true;
            var connections = _connections.Values.ToArray();
            Parallel.ForEach(connections, connection =>
            {
                connection.Close();
            });
            _processThread?.Join();
            _updateThread?.Join();
            _client.Close();
            _incomeConnections.TryReceiveAll(out _);
            _incomeConnections.Complete();
        }
    }
}
