﻿using System;
using System.Numerics;

namespace KcpNet
{
    internal abstract class KcpPacket
    {
        public KcpPacketType Type { get; }

        protected KcpPacket(KcpPacketType type)
        {
            Type = type;
        }

        public abstract byte[] GetBuffer();
    }



    internal class KcpCommand : KcpPacket, IEquatable<KcpCommand>
    {
        private readonly byte[] _buffer;

        public KcpCommand(KcpPacketType type) : base(type)
        {
            _buffer = new byte[1];
            _buffer[0] = (byte) Type;
        }

        public bool Equals(KcpCommand other)
        {
            if (other == null)
            {
                return false;
            }
            return Type == other.Type;
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((KcpCommand)obj);
        }

        public override int GetHashCode()
        {
            return (int)Type;
        }

        public static bool operator ==(KcpCommand left, KcpCommand right)
        {
            return Equals(left, right);
        }

        public static bool operator !=(KcpCommand left, KcpCommand right)
        {
            return !Equals(left, right);
        }

        public override byte[] GetBuffer()
        {
            return _buffer;
        }
    }

    internal class KcpPingRequest : KcpCommand
    {
        public static readonly KcpPingRequest Shared = new();

        public KcpPingRequest() : base(KcpPacketType.PingRequest)
        {
        }
    }


    internal class KcpPingResponse : KcpCommand
    {
        public static readonly KcpPingResponse Shared = new();

        public KcpPingResponse() : base(KcpPacketType.PingResponse)
        {
        }
    }


    internal class KcpClose : KcpCommand
    {
        public static readonly KcpClose Shared = new();

        public KcpClose() : base(KcpPacketType.Close)
        {
        }
    }


    internal class KcpData : KcpPacket, IEquatable<KcpData>
    {
        private readonly byte[] _buffer;
        private readonly Memory<byte> _dataBuffer;
        private readonly int _dataSize;

        private readonly int _hashCode;

        public int Offset { get; set; } =  0;

        public int Length => _dataSize - Offset;


        public KcpData(byte[] data) : base(KcpPacketType.Data)
        {
            _dataSize = data.Length;
            _buffer = new byte[5 + _dataSize];
            _buffer[0] = (byte) Type;
            Array.Copy(BitConverter.GetBytes(_dataSize),0,_buffer,1,4);
            Array.Copy(data, 0, _buffer, 5, _dataSize);
            _dataBuffer = new Memory<byte>(_buffer, 5, _dataSize);
            var bigInt = new BigInteger(_buffer);
            _hashCode = bigInt.GetHashCode();
        }

        public byte[] GetData()
        {
            return _dataBuffer.Slice(Offset).ToArray();
        }

        public override byte[] GetBuffer()
        {
            return _buffer;
        }

        public bool Equals(KcpData other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            return _hashCode == other._hashCode;
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != this.GetType()) return false;
            return Equals((KcpData)obj);
        }

        public override int GetHashCode()
        {
            return _hashCode;
        }

        public static bool operator ==(KcpData left, KcpData right)
        {
            return Equals(left, right);
        }

        public static bool operator !=(KcpData left, KcpData right)
        {
            return !Equals(left, right);
        }
    }
}
