﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Threading.Tasks.Dataflow;

namespace KcpNet
{
    internal enum PacketPriority
    {
        Low,
        Middle,
        High,
    }

    /// <summary>
    /// Store packet into this buffer, get packet with priority.
    /// </summary>
    internal class KcpPriorityPacketBuffer
    {
        private readonly BufferBlock<KcpPacket> _lowBuffer;
        private readonly BufferBlock<KcpPacket> _middleBuffer;
        private readonly BufferBlock<KcpPacket> _highBuffer;

        private readonly int _lowBufferLength;
        private readonly int _middleBufferLength;
        private readonly int _highBufferLength;

        public KcpPriorityPacketBuffer(int lowBufferLength = -1, int middleBufferLength = -1, int highBufferLength = -1)
        {
            _lowBufferLength = lowBufferLength;
            _middleBufferLength = middleBufferLength;
            _highBufferLength = highBufferLength;

            _lowBuffer = new BufferBlock<KcpPacket>();
            _middleBuffer = new BufferBlock<KcpPacket>();
            _highBuffer = new BufferBlock<KcpPacket>();
        }


        public async Task<bool> WritePacketAsync(KcpPacket packet, PacketPriority priority)
        {
           return await WritePacketAsync(packet, priority, CancellationToken.None).ConfigureAwait(false);
        }

        public async Task<bool> WritePacketAsync(KcpPacket packet, PacketPriority priority, CancellationToken cancellationToken)
        {
            bool result;
            switch (priority)
            {
                case PacketPriority.Low:
                    if (_lowBuffer.Count >= _lowBufferLength)
                    {
                        return false;
                    }
                    result = await _lowBuffer.SendAsync(packet, cancellationToken).ConfigureAwait(false);
                    break;
                case PacketPriority.Middle:
                    if (_middleBuffer.Count >= _middleBufferLength)
                    {
                        return false;
                    }
                    result = await _middleBuffer.SendAsync(packet, cancellationToken).ConfigureAwait(false);
                    break;
                case PacketPriority.High:
                    if (_highBuffer.Count >= _highBufferLength)
                    {
                        return false;
                    }
                    result = await _highBuffer.SendAsync(packet, cancellationToken).ConfigureAwait(false);
                    break;
                default:
                    throw new ArgumentOutOfRangeException(nameof(priority), priority, null);
            }
            return result;
        }


        public bool WritePacket(KcpPacket packet, PacketPriority priority)
        {
            bool result;
            switch (priority)
            {
                case PacketPriority.Low:
                    if (_lowBuffer.Count >= _lowBufferLength)
                    {
                        return false;
                    }
                    result = _lowBuffer.Post(packet);
                    break;
                case PacketPriority.Middle:
                    if (_middleBuffer.Count >= _middleBufferLength)
                    {
                        return false;
                    }
                    result = _middleBuffer.Post(packet);
                    break;
                case PacketPriority.High:
                    if (_highBuffer.Count >= _highBufferLength)
                    {
                        return false;
                    }
                    result = _highBuffer.Post(packet);
                    break;
                default:
                    throw new ArgumentOutOfRangeException(nameof(priority), priority, null);
            }

            return result;
        }


        public KcpPacket ReadPacket()
        {
            var packets = new List<KcpPacket>();
            if (_highBuffer.TryReceive(out var highPacket))
            {
                return highPacket;
            }
            if (_middleBuffer.TryReceive(out var middlePacket))
            {
                return middlePacket;
            }
            if (_lowBuffer.TryReceive(out var lowPacket))
            {
                return lowPacket;
            }

            return null;
        }

        public IList<KcpPacket> ReadPackets(PacketPriority priority, bool distinct = false)
        {
            var packets = new List<KcpPacket>();
            switch (priority)
            {
                case PacketPriority.Low:
                    if (_lowBuffer.TryReceiveAll(out var lowPackets))
                    {
                        if (distinct)
                        {
                            lowPackets = lowPackets.Distinct().ToList();
                        }
                        packets.AddRange(lowPackets);
                    }
                    break;
                case PacketPriority.Middle:
                    if (_middleBuffer.TryReceiveAll(out var middlePackets))
                    {
                        if (distinct)
                        {
                            middlePackets = middlePackets.Distinct().ToList();
                        }
                        packets.AddRange(middlePackets);
                    }
                    break;
                case PacketPriority.High:
                    if (_highBuffer.TryReceiveAll(out var highPackets))
                    {
                        if (distinct)
                        {
                            highPackets = highPackets.Distinct().ToList();
                        }
                        packets.AddRange(highPackets);
                    }
                    break;
            }

            return packets;
        }


        public IList<KcpPacket> ReadPackets(bool distinct = false)
        {
            var packets = new List<KcpPacket>();
            if (_highBuffer.TryReceiveAll(out var highPackets))
            {
                packets.AddRange(highPackets);
            }
            if (_middleBuffer.TryReceiveAll(out var middlePackets))
            {
                packets.AddRange(middlePackets);
            }
            if (_lowBuffer.TryReceiveAll(out var lowPackets))
            {
                packets.AddRange(lowPackets);
            }
            return distinct? packets.Distinct().ToList(): packets;
        }


        public void Close()
        {
            _lowBuffer.TryReceiveAll(out _);
            _middleBuffer.TryReceiveAll(out _);
            _highBuffer.TryReceiveAll(out _);
            _lowBuffer.Complete();
            _middleBuffer.Complete();
            _highBuffer.Complete();
        }
    }
}
